#############################################################################
# wordstat.py - Tools for language statistics
#
# This Python library contains various tools for statistical analysis of
# languages.
##############################################################################

import re, string




#----------------------------------------------------------------------------
# add - add an n-gram to the histogram
#
# Given a word and an n-gram extracted from it, this function increments
# the appropriate bin in a histogram of n-gram probabilities.  The word
# argument is used only to report errors.  The ngram argument is assumed
# to already have been lowercased and filtered to remove hyphens and
# apostrophes.
#----------------------------------------------------------------------------

def add(word, ngram, hist):
    
    c01 = ngram[0:-1]
    c2  = ngram[-1]
    if c2 == ' ':
        i = 0
    elif ord('a') <= ord(c2) and ord(c2) <= ord('z'):
        i = ord(c2) - ord('a') + 1
    else:
        print "Invalid word:", string.strip(word)
        return
    if not hist.has_key(c01):
        hist[c01] = 27 * [0]
    lst = hist[c01]
    lst[i] = lst[i] + 1



#-----------------------------------------------------------------------------
# scan - scan word for n-grams
#
# This function scans a word for n-grams and updates the n-gram histogram
# for each one found.  The nmark argument is the number of prefix characters,
# or one less than the N in n-grams to be computed.
#-----------------------------------------------------------------------------

def scan(word, nmark, hist):
    
    padded = ((nmark)*" ") + word + " "
    for i in range(len(padded) - nmark):
        ngram = padded[i:i+nmark+1]
        add(word, ngram, hist)



#-----------------------------------------------------------------------------
# ngram_prob - Compute n-gram probabilities from a set of word lists
#
# This function reads a set of files that contain word lists and computes
# the transition probabilities for an nth order Markov chain model.  It also
# generates a dictionary containing all the words read, which can be used
# to reject true words.
#
# Each wordlist file is assumed to contain one word per line.
#
# Only words that consist of only lowercase letters are considered to be
# proper English words and are used to compute the n-gram probabilities.
# This excludes capitalized words (presumably proper names) and words that
# contain digits, ampersands, and apostrophes.  Words containing hyphens are
# alse excluded, although this is probably not the right treatement.

# Two values are returned: A dictionary containing all the words seen in
# any of the word lists, and another dictionary giving the transition
# probabilities.  The transition probabilities are given as a hash keyed
# on the initial two letters; the value is a list containing the probability
# of the letters SP, a, b, ..., z following the initial letters.  The special
# keys "  x" and " xy" correspond to initial x or xy for trigraphs, and
# similarly for tetragraphs.
#
# FIXME: Think about the proper treatment of hyphenated words.
#-----------------------------------------------------------------------------

def ngram_prob(wordlists, nmark) :

    # Create empty prob table and word dictionary
    dict = {}
    prob = {}

    # Loop over all the word lists
    for wordlist in wordlists :

        # Read and process the current wordlist
        wordin = open(wordlist, "r")
        while (1):

            # Read the next word; stop at end of file
            word = wordin.readline();
            if not word:
                break

            # Strip leading and trailing whitespace
            word = string.strip(word)
            
            # Discard any word that contains a digit
            if re.search(r'\d', word) :
                continue

            # Generate dictionary key by lowercasing and removing odd chars
            key = string.lower(word)
            key = re.sub(r"[-' ]", "", key) 

            # Add to the dictionary, if not already there
            if not dict.has_key(key) :
                dict[key] = 1

            # Scan words containing only letters for n-grams
            if re.match(r'[a-z]+$', word) :
                scan(word, nmark, prob)

        wordin.close()

    # Post process counts into probabilities
    for prefix in prob.keys():
        count = 0.0
        lst = prob[prefix]
        for i in range(27):
            count = count + lst[i]
        for i in range(27):
            lst[i] = lst[i] / count

    # Return the dictionary and probabilities
    return dict, prob
